import defu from 'defu';

class DoclifyException extends Error {
  constructor(message, status, data) {
    if (typeof status === "undefined") {
      status = -1;
    }
    if (status >= 500) {
      message = "Unknown error occurred, please try again later.";
    } else if (data && data.error) {
      message = data.error.message;
    } else if (Array.isArray(data)) {
      message = "Validation failed. Please verify your paramaters.";
    }
    super(message);
    this.isDoclify = true;
    this.status = status;
    this.data = data;
    Object.setPrototypeOf(this, new.target.prototype);
  }
  get isValidationError() {
    var _a;
    return Array.isArray((_a = this.data) == null ? void 0 : _a.errors);
  }
  get validationErrors() {
    var _a, _b;
    return (_b = (_a = this.data) == null ? void 0 : _a.errors) != null ? _b : [];
  }
  get statusCode() {
    return this.status;
  }
  get code() {
    var _a, _b;
    return (_b = (_a = this.data) == null ? void 0 : _a.error) == null ? void 0 : _b.code;
  }
  toJSON() {
    return {
      message: this.message,
      status: this.status,
      code: this.code,
      data: this.data
    };
  }
  static fromJSON(json) {
    return new DoclifyException(json.message, json.status, json.data);
  }
  static fromResponse(response, data) {
    return new DoclifyException("Invalid request.", response.status, data);
  }
}

class Documents {
  constructor(client) {
    this.client = client;
    this._include = [];
    this._select = [];
    this._order = [];
    this.q = [];
    if (client.options.language) {
      this._lang = client.options.language;
    }
  }
  where(field, operator, value) {
    this.q.push([field, operator, value]);
    return this;
  }
  collection(value) {
    return this.eq("sys.collection", value);
  }
  contentType(value) {
    return this.eq("sys.contentType", value);
  }
  id(value) {
    return this.eq("sys.id", value);
  }
  uid(value) {
    return this.eq("sys.uid", value);
  }
  eq(field, value) {
    return this.where(field, "eq", value);
  }
  not(field, value) {
    return this.where(field, "not", value);
  }
  in(field, value) {
    return this.where(field, "in", value);
  }
  nin(field, value) {
    return this.where(field, "nin", value);
  }
  gt(field, value) {
    return this.where(field, "gt", value);
  }
  gte(field, value) {
    return this.where(field, "gte", value);
  }
  lt(field, value) {
    return this.where(field, "lt", value);
  }
  lte(field, value) {
    return this.where(field, "lte", value);
  }
  fulltext(field, value) {
    return this.where(field, "fulltext", value);
  }
  match(field, value) {
    return this.where(field, "match", value);
  }
  query(query) {
    query(this);
    return this;
  }
  language(lang) {
    this._lang = lang;
    return this;
  }
  limit(limit) {
    this._limit = limit;
    return this;
  }
  offset(offset) {
    this._offset = offset;
    return this;
  }
  include(fields) {
    this._include.push(...fields);
    return this;
  }
  select(fields) {
    this._select.push(...fields);
    return this;
  }
  orderBy(field, asc = "asc") {
    this._order.push([field, asc]);
    return this;
  }
  getParams(params = {}) {
    return Object.assign({
      q: JSON.stringify(this.q),
      include: this._include.length ? JSON.stringify(this._include) : void 0,
      order: this._order.length ? JSON.stringify(this._order) : void 0,
      select: this._select.length ? JSON.stringify(this._select) : void 0,
      lang: this._lang,
      limit: this._limit,
      offset: this._offset
    }, params);
  }
  fetch() {
    return this.client.request("documents/search", {
      params: this.getParams()
    });
  }
  paginate(page, perPage) {
    return this.client.request("documents/paginated", {
      params: this.getParams({
        perPage,
        page
      })
    });
  }
  async first() {
    const results = await this.client.request("documents/search", {
      params: this.getParams({
        limit: 1
      })
    });
    return results.length ? results[0] : null;
  }
  async firstOrFail() {
    const results = await this.client.request("documents/search", {
      params: this.getParams({
        limit: 1
      })
    });
    if (!results.length) {
      throw new DoclifyException("Document was not found.", 404);
    }
    return results[0];
  }
  then(resolve, reject) {
    return this.fetch().then(resolve, reject);
  }
}

const escapeHTML = function(text) {
  const htmlEscapes = {
    "&": "&amp;",
    "<": "&lt;",
    ">": "&gt;",
    '"': "&quot;",
    "'": "&#39;"
  };
  const reUnescapedHtml = /[&<>"']/g;
  const reHasUnescapedHtml = RegExp(reUnescapedHtml.source);
  if (text && reHasUnescapedHtml.test(text)) {
    return text.replace(reUnescapedHtml, (chr) => htmlEscapes[chr]);
  }
  return text;
};

const isEmailLinkType = (type) => type === "email";
const schema = {
  nodes: {
    horizontal_rule() {
      return {
        singleTag: "hr"
      };
    },
    blockquote() {
      return {
        tag: "blockquote"
      };
    },
    bullet_list() {
      return {
        tag: "ul"
      };
    },
    code_block(node) {
      return {
        tag: [
          "pre",
          {
            tag: "code",
            attrs: node.attrs
          }
        ]
      };
    },
    hard_break() {
      return {
        singleTag: "br"
      };
    },
    heading(node) {
      return {
        tag: `h${node.attrs.level}`
      };
    },
    image(node) {
      return {
        singleTag: [{
          tag: "img",
          attrs: {
            src: node.attrs.url,
            alt: node.attrs.name
          }
        }]
      };
    },
    list_item() {
      return {
        tag: "li"
      };
    },
    ordered_list() {
      return {
        tag: "ol"
      };
    },
    paragraph() {
      return {
        tag: "p"
      };
    },
    table() {
      return {
        tag: [
          "table",
          "tbody"
        ]
      };
    },
    table_row() {
      return { tag: "tr" };
    },
    table_header(node) {
      return {
        tag: [{ tag: "th", attrs: node.attrs }]
      };
    },
    table_cell(node) {
      return {
        tag: [{ tag: "td", attrs: node.attrs }]
      };
    }
  },
  marks: {
    bold() {
      return {
        tag: "b"
      };
    },
    strike() {
      return {
        tag: "strike"
      };
    },
    underline() {
      return {
        tag: "u"
      };
    },
    strong() {
      return {
        tag: "strong"
      };
    },
    code() {
      return {
        tag: "code"
      };
    },
    italic() {
      return {
        tag: "i"
      };
    },
    link(node) {
      const attrs = { ...node.attrs };
      const { linktype = "url" } = node.attrs;
      attrs.rel = "noopener";
      if (isEmailLinkType(linktype)) {
        attrs.href = `mailto:${attrs.href}`;
      }
      if (attrs.anchor) {
        attrs.href = `${attrs.href}#${attrs.anchor}`;
        delete attrs.anchor;
      }
      return {
        tag: [{
          tag: "a",
          attrs
        }]
      };
    },
    styled(node) {
      return {
        tag: [{
          tag: "span",
          attrs: node.attrs
        }]
      };
    }
  }
};

class HTMLSerializer {
  constructor(schema$1 = schema) {
    this.nodes = {};
    this.marks = {};
    this.marks = schema$1.marks;
    this.nodes = schema$1.nodes;
  }
  addNode(key, schema) {
    this.nodes[key] = schema;
  }
  addMark(key, schema) {
    this.marks[key] = schema;
  }
  render(data) {
    if (!data) {
      return "";
    }
    if (data.content && Array.isArray(data.content)) {
      const html = [];
      data.content.forEach((node) => {
        html.push(this.renderNode(node));
      });
      return html.join("");
    }
    console.warn("The render method must receive an object with a content field, which is an array");
    return "";
  }
  renderNode(item) {
    const html = [];
    if (item.marks) {
      item.marks.forEach((m) => {
        const mark = this.getMatchingMark(m);
        if (mark) {
          html.push(this.renderOpeningTag(mark.tag));
        }
      });
    }
    const node = this.getMatchingNode(item);
    if (node && node.tag) {
      html.push(this.renderOpeningTag(node.tag));
    }
    if (item.content) {
      item.content.forEach((content) => {
        html.push(this.renderNode(content));
      });
    } else if (item.text) {
      html.push(escapeHTML(item.text));
    } else if (node && node.singleTag) {
      html.push(this.renderTag(node.singleTag, " /"));
    } else if (node && node.html) {
      html.push(node.html);
    }
    if (node && node.tag) {
      html.push(this.renderClosingTag(node.tag));
    }
    if (item.marks) {
      item.marks.slice(0).reverse().forEach((m) => {
        const mark = this.getMatchingMark(m);
        if (mark) {
          html.push(this.renderClosingTag(mark.tag));
        }
      });
    }
    return html.join("");
  }
  renderTag(tags, ending) {
    if (tags.constructor === String) {
      return `<${tags}${ending}>`;
    }
    const all = tags.map((tag) => {
      if (tag.constructor === String) {
        return `<${tag}${ending}>`;
      } else {
        let h = `<${tag.tag}`;
        if (tag.attrs) {
          for (const key in tag.attrs) {
            const value = tag.attrs[key];
            if (value !== null) {
              h += ` ${key}="${value}"`;
            }
          }
        }
        return `${h}${ending}>`;
      }
    });
    return all.join("");
  }
  renderOpeningTag(tags) {
    return this.renderTag(tags, "");
  }
  renderClosingTag(tags) {
    if (tags.constructor === String) {
      return `</${tags}>`;
    }
    const all = tags.slice(0).reverse().map((tag) => {
      if (tag.constructor === String) {
        return `</${tag}>`;
      } else {
        return `</${tag.tag}>`;
      }
    });
    return all.join("");
  }
  getMatchingNode(item) {
    if (typeof this.nodes[item.type] !== "function") {
      return;
    }
    return this.nodes[item.type](item);
  }
  getMatchingMark(item) {
    if (typeof this.marks[item.type] !== "function") {
      return;
    }
    return this.marks[item.type](item);
  }
}

const defaultSerializer = new HTMLSerializer();
const asHTML = (doc) => {
  return defaultSerializer.render(doc);
};

var index = /*#__PURE__*/Object.freeze({
  __proto__: null,
  HTMLSerializer: HTMLSerializer,
  defaultSerializer: defaultSerializer,
  asHTML: asHTML
});

var DOM = /*#__PURE__*/Object.freeze({
  __proto__: null,
  richText: index
});

const defaults = {
  timeout: 5e3
};
class Doclify {
  constructor(options = {}) {
    this.dom = DOM;
    this.options = defu(options || {}, defaults);
    if (!this.options.url) {
      if (!this.options.repository) {
        throw new TypeError("Repository or URL option is required.");
      }
      if (!this.options.key) {
        throw new TypeError("API key is required.");
      }
    }
    if (typeof options.fetch === "function") {
      this.fetch = options.fetch;
    } else if (typeof globalThis.fetch === "function") {
      this.fetch = globalThis.fetch;
    } else {
      throw new DoclifyException("A valid fetch implementation was not provided. In environments where fetch is not available (including Node.js), a fetch implementation must be provided via a polyfill or the `fetch` option.");
    }
    if (this.fetch === globalThis.fetch) {
      this.fetch = this.fetch.bind(globalThis);
    }
  }
  get baseUrl() {
    var _a;
    return this.options.url || `https://${(_a = this.options.repository) != null ? _a : ""}.cdn.doclify.io/api/v2`;
  }
  async request(endpoint, options = {}) {
    options.headers = options.headers || {};
    options.params = options.params || {};
    if (this.options.key) {
      options.headers.Authorization = "Bearer " + this.options.key;
    }
    if (this.options.language && !options.params.lang) {
      options.params.lang = this.options.language;
    }
    const searchParams = new URLSearchParams();
    for (const key in options.params) {
      if (options.params[key]) {
        searchParams.append(key, String(options.params[key]));
      }
    }
    delete options.params;
    const query = searchParams.toString();
    let response;
    try {
      response = await this.fetch(this.baseUrl + "/" + endpoint + (query ? "?" + query : ""), options);
    } catch (err) {
      throw new DoclifyException("Networking issue: " + err.message, 500);
    }
    const json = await response.json();
    if (!response.ok) {
      throw DoclifyException.fromResponse(response, json);
    }
    return json;
  }
  documents() {
    return new Documents(this);
  }
}

export { DOM, Doclify as default };
